/*************************************************************************
*
*	File: bmp.cpp
*	Copyright(c) Andreas Magnusson 2002
*
*	This is my file for reading and writing bmp-files.
*
*	Should work on both big-endian (eg. Sun Sparc) and little-endian (eg. Intel x86)
*	machines.
*
*	DISCLAIMER AND STUFF:
*	This source-code is supplied AS IS. I cannot be held responsible for any damage caused
*	by use or misuse of this source-code nor any derivatives of it.
*	So use the code as you please, but give credit where they are due.
*
*************************************************************************/

// disable the annoying warnings that comes up when you use map...
#pragma warning( disable : 4786)
#include "image.h"
#include <stdio.h>

//////////////////////////////////

struct bmp_header
{
	char bm[2];
	unsigned long file_size;
	unsigned long reserved1;
	unsigned long data_offset;
	unsigned long bmp_header_size;
	unsigned long width;
	unsigned long height;
	unsigned short planes;
	unsigned short bpp;
	unsigned long compr;
	unsigned long bmp_data_size;
	unsigned long hres;
	unsigned long vres;
	unsigned long colors;
	unsigned long imp_colors;
	//
	unsigned long make_long(unsigned char *&buf)
	{
		unsigned long rv = *buf++;
		rv += (*buf++ << 8);
		rv += (*buf++ << 16);
		rv += (*buf++ << 24);
		return rv;
	}
	unsigned short make_word(unsigned char *&buf)
	{
		unsigned short rv = *buf++;
		rv += (*buf++ << 8);
		return rv;
	}
	void long2byte(unsigned long v, unsigned char *&buf)
	{
		*buf++ = v & 0xff;
		*buf++ = (v >> 8) & 0xff;
		*buf++ = (v >> 16) & 0xff;
		*buf++ = (v >> 24) & 0xff;
	}
	void word2byte(unsigned short v, unsigned char *&buf)
	{
		*buf++ = v & 0xff;
		*buf++ = (v >> 8) & 0xff;
	}
	bmp_header(unsigned char *temphead)
	{
	  // this procedure is needed for compatibility between
		// little-endian machines (eg Intel-based) and big-endian
		// machines (eg Sun Sparc)
		bm[0] = *temphead++;
		bm[1] = *temphead++;
		file_size = make_long(temphead);
		reserved1 = make_long(temphead);
		data_offset = make_long(temphead);
		bmp_header_size = make_long(temphead);
		width = make_long(temphead);
		height = make_long(temphead);
		planes = make_word(temphead);
		bpp = make_word(temphead);
		compr = make_long(temphead);
		bmp_data_size = make_long(temphead);
		hres = make_long(temphead);
		vres = make_long(temphead);
		colors = make_long(temphead);
		imp_colors = make_long(temphead);
		//
	}
	bmp_header()
	{
	}
	void copy_to_buffer(unsigned char *tempbuf)
	{
		*tempbuf++ = 'B'; //bm[0];
		*tempbuf++ = 'M'; //bm[1];
		long2byte(file_size, tempbuf);
		long2byte(reserved1, tempbuf);
		long2byte(data_offset, tempbuf);
		long2byte(bmp_header_size, tempbuf);
		long2byte(width, tempbuf);
		long2byte(height, tempbuf);
		word2byte(planes, tempbuf);
		word2byte(bpp, tempbuf);
		long2byte(compr, tempbuf);
		long2byte(bmp_data_size, tempbuf);
		long2byte(hres, tempbuf);
		long2byte(vres, tempbuf);
		long2byte(colors, tempbuf);
		long2byte(imp_colors, tempbuf);
	}
};

///////////////////////////////////////////////////////////////////////////////////

class bmp_image : public image
{
	void bgr2rgb();
	void rgb2bgr();
public:
	bmp_image();
	virtual bool load(const std::string &file_name);
	virtual bool save(const std::string &file_name);
	virtual const char *get_type() {return "bmp";}
};

// register bmp_image to image factory
namespace
{
	register_in_factory<image, bmp_image> register_me("bmp");
};

bmp_image::bmp_image()
{
}

bool bmp_image::load(const std::string &file_name)
{
	FILE *file;
	
  file = fopen(file_name.c_str(), "rb");
	if(!file)
    return false;

	unsigned char *temphead = new unsigned char[sizeof(bmp_header)];
	fread(temphead, 1, sizeof(bmp_header), file);
	bmp_header header(temphead);
	delete[] temphead;

	if(strncmp(header.bm, "BM", 2) != 0)
	{
		fclose(file);
		return false;
	}

	// currently only handles 24 and 32 bit images
	if(header.bpp < 24)
	{
		fclose(file);
		return false;
	}

	width = header.width;
	height = header.height;

	color_depth = header.bpp;

	int ch_count = get_color_depth() >> 3;

	long size = width * height * ch_count;

	if(buffer)
		delete[] buffer;
  
	buffer = new unsigned char[size];
	if(buffer == NULL)
	{
		fclose(file);
		return false;
	}

	if(ftell(file) != header.data_offset)
		fseek(file, header.data_offset, SEEK_SET);

	// the bmp-image is read into the buffer backwards, since it seems to be stored
	// upside down...
	int oneline = get_width() * ch_count;
	int offset = size;
	for(int i = 0; i < size; i += oneline)
	{
		fread(&buffer[offset - oneline], 1, oneline, file);
		offset -= oneline;
	}
	fclose(file);

	// convert to rgb
	bgr2rgb();

	return true;
}

bool bmp_image::save(const std::string &file_name)
{
	bmp_header header;
	header.bpp = get_color_depth();
	header.colors = 0;
	header.imp_colors = 0;
	header.compr = 0;
	header.height = get_height();
	header.width = get_width();
	header.planes = 1;
	header.bmp_data_size = get_width() * get_height() * (get_color_depth() >> 3);
	header.bmp_header_size = 0x28;
	header.data_offset = sizeof(bmp_header);
	header.file_size = header.bmp_data_size + header.bmp_header_size;
	header.hres = 2834;
	header.vres = 2834;
	header.reserved1 = 0;
	//
	unsigned char *temphead = new unsigned char[sizeof(bmp_header)];
	if(!temphead)
		return false;

	FILE *fp = fopen(file_name.c_str(), "wb");
	if(!fp)
	{
		delete[] temphead;
		return false;
	}

	header.copy_to_buffer(temphead);
	fwrite(temphead, 1, sizeof(bmp_header), fp);

	// convert from rgb to bgr (since bmp seems to be that way)
	bgr2rgb();

	// write the image backwards, since that seems to be right...
	int oneline = get_width() * (get_color_depth() >> 3);
	unsigned long size = oneline * get_height();
	int offset = size;
	for(int i = 0; i < size; i += oneline)
	{
		fwrite(&buffer[offset - oneline], 1, oneline, fp);
		offset -= oneline;
	}
	fclose(fp);
	delete[] temphead;

	// convert the buffer back to rgb, since that is standard in my library.
	bgr2rgb();
	return true;
}

///
// this routine swaps the blue and red components since bmp seems to be in BGR format
// instead of rgb.
void bmp_image::bgr2rgb()
{
	const int color_planes = get_color_depth() >> 3;
	int size = get_width() * get_height() * color_planes;
	unsigned char tmp;
	for(int i = 0; i < size; i += color_planes)
	{
		tmp = buffer[i];
		buffer[i] = buffer[i + 2];
		buffer[i + 2] = tmp;
	}
}
